import numpy as np
from decimal import Decimal

from .magnitude import Magnitude

class UnitType:

    def __new__(cls, baseunits1, baseunits2):
        obj = object.__new__(cls)
        obj.baseunits1 = baseunits1
        obj.baseunits2 = baseunits2
        if obj._istype():
            return obj
        else:
            return None
            
    def convert(self, magnitude1):
        if not hasattr(self, self.conversion[0]):
                raise Exception('Conversion method is not implemented:', self.conversion[0])
        if isinstance(magnitude1.value, Decimal) or \
           isinstance(self.baseunits1.magnitude, Decimal) or \
           isinstance(self.baseunits2.magnitude, Decimal):
            magnitude1.value = Decimal(magnitude1.value)
            self.baseunits1.magnitude = Decimal(self.baseunits1.magnitude)
            self.baseunits2.magnitude = Decimal(self.baseunits2.magnitude)
        return Magnitude(
            getattr(self, self.conversion[0])(magnitude1.value * self.baseunits1.magnitude, *self.conversion[1:]) / self.baseunits2.magnitude,
            magnitude1.error
        )
        
    def add(self, unit1, unit2):
        if self.baseunits1.dimensions!=self.baseunits2.dimensions:
            raise Exception('Only units with the same dimension can added together', unit1, unit2)
        return unit1.magnitude + unit2.to(unit1.baseunits).magnitude

    def sub(self, unit1, unit2):
        if self.baseunits1.dimensions!=self.baseunits2.dimensions:
            raise Exception('Only units with the same dimension can added together', unit1, unit2)
        return unit1.magnitude - unit2.to(unit1.baseunits).magnitude

class StandardUnitType(UnitType):

    def _istype(self):
        if self.baseunits1.dimensions==self.baseunits2.dimensions:
            self.conversion = (f"_convert_linear",)
        elif -self.baseunits1.dimensions==self.baseunits2.dimensions:
            self.conversion = (f"_convert_inversed",)
        elif self.baseunits1.nobase and self.baseunits2.units==['rad']:
            self.conversion = (f"_convert_linear",)
        else:
            return False
        return True

    def _convert_inversed(self, value):
        return 1/value

    def _convert_linear(self, value):
        return value

class TemperatureUnitType(UnitType):
    
    process = ['Cel','degF']
    
    def _istype(self):
        if np.any(np.in1d(self.baseunits1.units+self.baseunits2.units, self.process)):
            if len(self.baseunits1.units)!=1 or len(self.baseunits2.units)!=1:
                raise Exception("Only simple units can be converted between each other:",
                                self.baseunits1.units, self.baseunits2.units)
            self.conversion = (f"_convert_{self.baseunits1.units[0]}_{self.baseunits2.units[0]}",)
        else:
            return False
        return True
    
    def _convert_K_Cel(self, value):
        return value-273.15
        
    def _convert_K_degF(self, value):
        return (value-273.15)*9/5+32
        
    def _convert_degR_degF(self, value):
        return value*9/5-459.67
        
    def _convert_degR_Cel(self, value):
        return (value*9/5-491.67)*5/9
        
    def _convert_Cel_K(self, value):
        return value+273.15
        
    def _convert_Cel_degF(self, value):
        return (value*9/5)+32
        
    def _convert_Cel_degR(self, value):
        return ((value*9/5)+491.67)*5/9
        
    def _convert_degF_K(self, value):
        return ((value-32)*5/9)+273.15
    
    def _convert_degF_Cel(self, value):
        return (value-32)*5/9
        
    def _convert_degF_degR(self, value):
        return (value+459.67)*5/9
        
        
class LogarithmicUnitType(UnitType):

    process: list = ['Np','B','Bm','BmW','BW','BV','BuV','BA','BuA','BOhm','BSPL','BSIL','BSWL']
    conversions: dict = {
        # power ratios
        'PR_Np':    ("_convert_Ratio_Np",  0.5, 1),    # Nepers
        'Np_PR':    ("_convert_Np_Ratio",  0.5, 1),
        'PR_B':     ("_convert_Ratio_B",   1,   1),    # Bel
        'B_PR':     ("_convert_B_Ratio",   1,   1),
        'W_Bm':     ("_convert_Ratio_B",   1,   1),
        'W_BmW':    ("_convert_Ratio_B",   1,   1),
        'Bm_W':     ("_convert_B_Ratio",   1,   1),
        'BmW_W':    ("_convert_B_Ratio",   1,   1),
        'W_BW':     ("_convert_Ratio_B",   1,   1e-3),
        'BW_W':     ("_convert_B_Ratio",   1,   1e3),
        'W_BSIL':   ("_convert_Ratio_B",   1,   1e9),  # W/m2
        'BSIL_W':   ("_convert_B_Ratio",   1,   1e-9),
        'W_BSWL':   ("_convert_Ratio_B",   1,   1e9),  # W
        'BSWL_W':   ("_convert_B_Ratio",   1,   1e-9),
        # amplitude ratios
        'AR_Np':    ("_convert_Ratio_Np",  1,   1),    # Nepers
        'Np_AR':    ("_convert_Np_Ratio",  1,   1),
        'AR_B':     ("_convert_Ratio_B",   2,   1),    # Bel
        'B_AR':     ("_convert_B_Ratio",   2,   1),
        'V_BV':     ("_convert_Ratio_B",   2,   1e-3),
        'BV_V':     ("_convert_B_Ratio",   2,   1e3),
        'V_BuV':    ("_convert_Ratio_B",   2,   1e3),
        'BuV_V':    ("_convert_B_Ratio",   2,   1e-3),
        'A_BA':     ("_convert_Ratio_B",   2,   1),
        'BA_A':     ("_convert_B_Ratio",   2,   1),
        'A_BuA':    ("_convert_Ratio_B",   2,   1e6),
        'BuA_A':    ("_convert_B_Ratio",   2,   1e-6),
        'Ohm_BOhm': ("_convert_Ratio_B",   2,   1e-3),
        'BOhm_Ohm': ("_convert_B_Ratio",   2,   1e3),
        'Pa_BSPL':  ("_convert_Ratio_B",   2,   50),
        'BSPL_Pa':  ("_convert_B_Ratio",   2,   0.02),
        # same decibel conversions
        'Bm_Bm':    ("_convert_B_B",       0),
        'BW_BW':    ("_convert_B_B",       0),
        'BmW_BmW':  ("_convert_B_B",       0),
        'BV_BV':    ("_convert_B_B",       0),
        'BuV_BuV':  ("_convert_B_B",       0),
        'BA_BA':    ("_convert_B_B",       0),
        'BuA_BuA':  ("_convert_B_B",       0),
        'BOhm_BOhm':("_convert_B_B",       0),
        'BSPL_BSPL':("_convert_B_B",       0),
        'BSIL_BSIL':("_convert_B_B",       0),
        'BSWL_BSWL':("_convert_B_B",       0),
        # different decibel conversions
        'BW_Bm':    ("_convert_B_B",       3),
        'Bm_BW':    ("_convert_B_B",      -3),
        'BW_BmW':   ("_convert_B_B",       3),
        'BmW_BW':   ("_convert_B_B",      -3),
        'Bm_BmW':   ("_convert_B_B",       0),
        'BmW_Bm':   ("_convert_B_B",       0),
        'BV_BuV':   ("_convert_B_B",       12),
        'BuV_BV':   ("_convert_B_B",      -12),
    }
    
    def _istype(self):
        if np.any(np.in1d(self.baseunits1.units+self.baseunits2.units, self.process)):
            if len(self.baseunits1.units) not in [1,2] or len(self.baseunits2.units) not in [1,2]:
                raise Exception("Only simple and fraction units can be converted between each other:",
                                self.baseunits1.units, self.baseunits2.units)
            conversion = f"{self.baseunits1.units[0]}_{self.baseunits2.units[0]}"
            if conversion in self.conversions:
                self.conversion = self.conversions[conversion]
            else:
                self.conversion = (f"_convert_{conversion}",)
        else:
            return False
        return True
    
    def _convert_B_B(self, value, exp=0):
        return value + exp
        
    def _convert_B_Np(self, value):
        return 1.151277918*value
        
    def _convert_Np_B(self, value):
        return value/1.151277918
        
    def _convert_Ratio_B(self, value, exp, conv):
        return exp*np.log10(value*conv)
        
    def _convert_B_Ratio(self, value, exp, conv):
        return np.power(10,value/exp)*conv
        
    def _convert_Ratio_Np(self, value, exp, conv):
        return exp*np.log(value*conv)
        
    def _convert_Np_Ratio(self, value, exp, conv):
        return np.exp(value/exp)*conv
        
    def add(self, unit1, unit2):
        if self.baseunits1.dimensions!=self.baseunits2.dimensions:
            raise Exception('Only units with the same dimension can added together', unit1, unit2)
        if self.baseunits1.units!=self.baseunits2.units:
            raise Exception('Only the same units can be added', unit1, unit2)
        mag1 = unit1.magnitude
        mag2 = unit2.to(unit1.baseunits).magnitude
        mag1.value = np.power(10,mag1.value*unit1.baseunits.magnitude)
        mag2.value = np.power(10,mag2.value*unit2.baseunits.magnitude)
        mag = mag1 + mag2
        mag.value = np.log10(mag.value)/unit1.baseunits.magnitude
        return mag

    def sub(self, unit1, unit2):
        if self.baseunits1.dimensions!=self.baseunits2.dimensions:
            raise Exception('Only units with the same dimension can added together', unit1, unit2)
        if self.baseunits1.units!=self.baseunits2.units:
            raise Exception('Only the same units can be substracted', unit1, unit2)
        mag1 = unit1.magnitude
        mag2 = unit2.to(unit1.baseunits).magnitude
        mag1.value = np.power(10,mag1.value*unit1.baseunits.magnitude)
        mag2.value = np.power(10,mag2.value*unit2.baseunits.magnitude)
        mag = mag1 - mag2
        mag.value = np.log10(mag.value)/unit1.baseunits.magnitude
        return mag



import json

from .node_base import BaseNode

class TagsNode(BaseNode):
    keyword: str = 'tags'

    @staticmethod
    def is_node(parser):
        parser.kwd_tags()
        if parser.is_parsed('kwd_tags'):
            parser.part_value()
            parser.part_comment()
            return TagsNode(parser)
            
    def parse(self, env):
        if env.nodes[-1].keyword not in ['str','int','float','bool']:
            raise Exception("Format can be set only to str, int, float and bool nodes:", env.nodes[-1].code)
        tags = json.loads(self.value_raw)
        if not isinstance(tags, list):
            raise Exception("Tags can be input only as an array of strings, instead received:", tags)
        if env.nodes[-1].tags == None:
            env.nodes[-1].tags = tags
        else:
            env.nodes[-1].tags += tags
        return None



import sys, os
import pytest
import numpy as np
sys.path.insert(0, 'src')

from scinumtools.units import Quantity
from scinumtools.dip import DIP, Environment
from scinumtools.dip.settings import Format
from scinumtools.dip.datatypes import IntegerType, FloatType, StringType, BooleanType
from scinumtools.dip.solvers import TemplateSolver

@pytest.fixture
def test_path():
    return "examples/"

def test_example_of_use(test_path):

    with DIP() as dip:
        dip.from_string("""
        mpi
          nodes int = 36
          cores int = 96
        """)                               # get code from a string
        env1 = dip.parse()                 # parse the code

    with DIP(env1) as dip:                 # pass environment to a new DIP instance
        dip.from_file(test_path+"settings2.dip") # add new parameter
        env2 = dip.parse()                 # parse new parameters

    nodes = env2.query("mpi.*")            # select nodes using a query
    geom = env2.request("?box.geometry")   # select a node using a request
    
    assert nodes[0].value == IntegerType(36)
    assert nodes[1].value == IntegerType(96)
    assert geom[0].value == IntegerType(3)

    data = env2.data(verbose=True)
    np.testing.assert_equal(data,{
        'mpi.nodes': 36,
        'mpi.cores': 96,
        'runtime.t_max': 10,
        'runtime.timestep': 0.01,
        'box.geometry': 3,
        'box.size.x': 10,
        'box.size.y': 3e7,
        'modules.heating': False,
        'modules.radiation': True,
    })

    data = env2.data(Format.TUPLE, verbose=True)
    np.testing.assert_equal(data,{
        'mpi.nodes': 36,
        'mpi.cores': 96,
        'runtime.t_max': (10, 'ns'),
        'runtime.timestep': (0.01, 'ns'),
        'box.geometry': 3,
        'box.size.x': (10, 'nm'),
        'box.size.y': (3e7,'nm'),
        'modules.heating': False,
        'modules.radiation': True,
    })
    
    data = env2.data(Format.QUANTITY, verbose=True)
    np.testing.assert_equal(data,{
        'mpi.nodes':         Quantity(36),
        'mpi.cores':         Quantity(96),
        'runtime.t_max':     Quantity(10, 'ns'),
        'runtime.timestep':  Quantity(0.01, 'ns'),
        'box.geometry':      Quantity(3),
        'box.size.x':        Quantity(10, 'nm'),
        'box.size.y':        Quantity(3e7, 'nm'),
        'modules.heating':   False,
        'modules.radiation': True,
    })
    
    data = env2.data(Format.TYPE, verbose=True)
    np.testing.assert_equal(data,{
        'mpi.nodes':         IntegerType(36),
        'mpi.cores':         IntegerType(96),
        'runtime.t_max':     FloatType(10, 'ns'),
        'runtime.timestep':  FloatType(0.01, 'ns'),
        'box.geometry':      IntegerType(3),
        'box.size.x':        FloatType(10, 'nm'),
        'box.size.y':        FloatType(3e7, 'nm'),
        'modules.heating':   BooleanType(False),
        'modules.radiation': BooleanType(True),
    })
    
def test_query_request_tag(test_path):
    
    with DIP() as dip:
        dip.from_string("""
        mpi
          nodes int = 36
          cores int = 96
        """)                               
        dip.from_file(test_path+"settings2.dip") 
        env = dip.parse()               
    
    nodes = env.query("mpi.*")
    assert len(nodes) == 2
    nodes = env.query("runtime.*", tags=['step'])  
    assert len(nodes) == 1
    geom = env.request("?box.geometry") 
    assert len(nodes) == 1
    geom = env.request("?runtime.*", tags=['step'])  
    assert len(nodes) == 1

    
def test_data_tag(test_path):
    
    with DIP() as dip:
        dip.from_string("""
        mpi
          nodes int = 36
          cores int = 96
        """)                               
        dip.from_file(test_path+"settings2.dip") 
        env = dip.parse()                 # parse the code
    
    data = env.data(query="mpi.*")
    np.testing.assert_equal(data,{
        'nodes': 36,
        'cores': 96,
    })
    data = env.data(tags=['step'])       
    np.testing.assert_equal(data,{
        'runtime.timestep': 0.01,
    })         
    data = env.data(query="mpi.*", tags=['step']) 
    np.testing.assert_equal(data,{})         


def test_definition_template(test_path):
    
    with DIP() as dip:
        dip.from_file(test_path+'definitions.dip')
        env3 = dip.parse()
        data = env3.data(format=Format.TYPE)
    np.testing.assert_equal(data,{
        'runtime.t_max':        FloatType(1e-08, 's'),
        'runtime.timestep':     FloatType(1.0000000000000001e-11, 's'),
        'box.geometry':         IntegerType(3),
        'box.size.x':           FloatType(1e-06, 'cm'),
        'box.size.y':           FloatType(3.0, 'cm'),
        'box.size.z':           FloatType(23.0, 'cm'),
        'modules.hydrdynamics': BooleanType(True),
        'modules.heating':      BooleanType(False),
        'modules.radiation':    BooleanType(True)
    })
    with TemplateSolver(env3) as ts:
        text = ts.template(test_path+'template.txt', test_path+'processed.txt')
    assert text == "Geometry: 3\nBox size: [1e-06, 3.0, 23.0]\n"

def test_units_sources(test_path):

    with DIP() as dip:
        dip.add_source("settings", test_path+'settings.dip')
        dip.add_unit("length", 1, "m")
        dip.from_string("""
        width float = 23 [length]
        x_size float = {settings?box.size.x}
        """)
        env = dip.parse()
        data = env.data(format=Format.TYPE)
    np.testing.assert_equal(data,{
        'width':  FloatType(23, '[length]'),
        'x_size': FloatType(10, 'nm'),
    })
        
def test_functions():

    def fn_volume(data):
        side = data['side'].convert('cm').value
        return side**3

    def fn_surface(data):
        side = data['side'].convert('cm').value
        return IntegerType(6*side**2, 'cm2')
    
    def is_prime(data):
        side = data['side'].convert('cm').value
        return side in [1,2,3,5,7,11]

    def print_value(data):
        return str(data['side'])
    
    with DIP() as dip:
        dip.add_function("fn_volume", fn_volume)
        dip.add_function("fn_surface", fn_surface)
        dip.add_function("is_prime", is_prime)
        dip.add_function("print_value", print_value)
        dip.from_string("""
        side float = 5 cm
        volume float = (fn_volume) cm3
        surface int = (fn_surface) mm2
        prime bool = (is_prime)
        value str = (print_value)
        """)
        env = dip.parse()
        data = env.data(format=Format.TYPE)
        np.testing.assert_equal(data,{
            'side':    FloatType(5, 'cm'),
            'volume':  FloatType(125, 'cm3'),
            'surface': IntegerType(15000, 'mm2'),
            'prime':   BooleanType(True),
            'value':   StringType("FloatType(5.0 cm)")
        })



import numpy as np
from math import isclose

from ...units import Quantity, UnitEnvironment
from .type import Type
from .type_boolean import BooleanType
from ..settings import Numeric

class NumberType(Type):
    dtype = None
    typename: str = 'number'

    def _prepare(self, other):
        if self.dtype not in [int,float,str,bool]:
            # if self node datatype is unknown
            if other.dtype in [int,float]:
                self.convert(other.unit)
            if other.dtype is None:
                self.value = float(self.value)
            else:
                self.value = other.dtype(self.value)
        elif other.dtype not in [int,float,str,bool]:
            # if other node datatype is unknown
            if self.dtype in [int,float]:
                other.convert(self.unit)
            other.value = self.dtype(other.value)
        elif type(self)==type(other):
            # if both datatypes are known
            if self.dtype in [int,float]:
                self.convert(other.unit)
        else:                               # throw error if both datatypes are unknown
            raise Exception("Invalid comparison:", expr)
        return self.value, other.value
    
    def __eq__(self, other):
        left, right = self._prepare(other)
        if isinstance(left,str) and isinstance(right,str):
            return BooleanType(left==right)
        else:
            return BooleanType(np.isclose(left, right, rtol=Numeric.PRECISION))
        
    def __ne__(self, other):
        left, right = self._prepare(other)
        return BooleanType(left != right)

    def __lt__(self, other):
        left, right = self._prepare(other)
        return BooleanType(left < right)

    def __gt__(self, other):
        left, right = self._prepare(other)
        return BooleanType(left > right)

    def __le__(self, other):
        left, right = self._prepare(other)
        return BooleanType((left<right)|np.isclose(left, right, rtol=Numeric.PRECISION))

    def __ge__(self, other):
        left, right = self._prepare(other)
        return BooleanType((left>right)|np.isclose(left, right, rtol=Numeric.PRECISION))
        
    def convert(self, unit, env=None):
        """ Convert units of this type
        """
        if unit:
            if self.unit and self.unit!=unit:
                if env is None:
                    self.value = Quantity(float(self.value), self.unit).value(unit)
                else:
                    with UnitEnvironment(env.units):
                        self.value = Quantity(float(self.value), self.unit).value(unit)
                self.unit = unit
        return self
 


import numpy as np
import re
import os
from pathlib import Path
from typing import List, Callable, Tuple
from inspect import getframeinfo, stack

from .environment import Environment
from .source import Source
from .settings import Keyword, Sign
from .nodes.parser import Parser
from .nodes import EmptyNode, ImportNode, UnitNode, SourceNode, CaseNode
from .nodes import OptionNode, ConstantNode, FormatNode, ConditionNode, TagsNode
from .nodes import ModNode, GroupNode
from .nodes import BooleanNode, IntegerNode, FloatNode, StringNode, TableNode
from .solvers import LogicalSolver
from .datatypes import Type

class DIP:
    """ DIP parser class

    :param str code: DIP code
    :param DIP_Environment env: DIP environment object
    """
    env: Environment
    lines: List[dict]
    source: Source

    def __init__(self, env:Environment=None, **kwargs):
        self.lines = []
        if env:
            self.env = env
        elif 'env' not in kwargs:
            self.env = Environment()
        else:
            self.env = None
        if 'source' not in kwargs:
            # determine which file instantiate this class
            caller = getframeinfo(stack()[1][0])
            if caller.filename == '<stdin>':                
                self.source = Source(lineno=caller.lineno, filename=os.getcwd())
            else:
                self.source = Source(lineno=caller.lineno, filename=caller.filename)
        else:
            self.source = kwargs['source']
        if 'docs' in kwargs:
            self.env.docs = kwargs['docs']
            del kwargs['docs']

    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        pass
    
    def from_file(self, filepath:str):
        """ Load DIP code from a file

        :param str filepath: Path to a DIP file
        """
        if not os.path.isabs(filepath):
            # set relative paths with respect to the source script
            parent = Path(self.source.filename)
            if os.path.isfile(parent):
                filepath = parent.parent / filepath
            else:
                filepath = parent / filepath
        with open(filepath,'r') as f:           
            lines = f.read().split('\n')
        for line,code in enumerate(lines):
            self.lines.append(dict(
                code = code,
                source = Source(
                    lineno=line+1,
                    filename=os.path.realpath(filepath),
                    primary=self.source.primary
                )
            ))
        
    def from_string(self, code:str):
        """ Use DIP code from a string

        :param str code: DIP code
        """
        lines = code.split('\n')
        for line,code in enumerate(lines):
            self.lines.append(dict(
                code = code,
                source = self.source,
            ))

    def add_source(self, name:str, path:str):
        self.lines.append(dict(
            code = f"{Sign.VARIABLE}{Keyword.SOURCE} {name} = {path}",
            source = self.source
        ))
        
    def add_unit(self, name:str, value:float, unit:str=None):
        if unit:
            code = f"{Sign.VARIABLE}{Keyword.UNIT} {name} = {value} {unit}"
        else:
            code = f"{Sign.VARIABLE}{Keyword.UNIT} {name} = {value}"
        self.lines.append(dict(
            code = code,
            source = self.source
        ))

    def add_function(self, name:str, fn:Callable):
        self.env.functions[name] = fn

    def parse(self):
        """ Parse DIP nodes from code lines
        """
        # Convert code lines to nodes
        env = Environment()
        while len(self.lines)>0:
            line = self.lines.pop(0)
            # Group block structures
            if '"""' in line['code']:
                block = []
                while len(self.lines)>0:
                    subline = self.lines.pop(0)
                    if '"""' in subline['code']:
                        line['code'] += "\n".join(block) + subline['code'].lstrip()
                        break
                    else:
                        block.append( subline['code'] )
                else:
                    raise Exception("Block structure starting on line %d is not properly terminated."%l)
            node = self._determine_node(line)
            env.nodes.append(node)
        # Parse nodes
        while len(env.nodes):
            node = env.pop_node()
            # Perform specific node parsing only outside of case or inside of valid case
            if self.env.is_case():
                node.inject_value(self.env)
                parsed = node.parse(self.env)
                if parsed: 
                    # Add parsed nodes to the queue and continue
                    env.prepend_nodes(parsed)
                    continue
            # Create hierarchical name
            excluded = ['empty','unit','source',
                        'option','constant','format','condition','tags']
            self.env.update_hierarchy(node, excluded)
            # Add nodes to the list
            excluded += ['group']
            if node.keyword in excluded:
                continue
            elif node.keyword=='case':   # Parse cases
                self.env.solve_case(node)
            else:
                if not self.env.is_case():
                    continue
                self.env.prepare_node(node)
                # Set the node value
                node.set_value()
                # If node was previously defined, modify its value
                for n in range(len(self.env.nodes)):
                    if self.env.nodes[n].name==node.name:
                        if self.env.nodes[n].constant:
                            raise Exception(f"Node '{self.env.nodes[n].name}' is constant and cannot be modified:",node.code)
                        if not self.env.docs:
                            self.env.nodes[n].modify_value(node, self.env)
                        break
                # If node wasn't defined, create a new node
                else:
                    if node.keyword=='mod' and node.source.primary:
                        raise Exception(f"Modifying undefined node:",node.name)
                    self.env.nodes.append(node)
        # Validate nodes
        for node in self.env.nodes:
            if not self.env.docs:
                # Check if all declared nodes have assigned value
                if node.defined and node.value is None:
                    raise Exception(f"Node value must be defined:", node.code)
                # check if node value is in options
                if isinstance(node,(IntegerNode, FloatNode, StringNode)):
                    node.validate_options()
                # Check conditions
                if node.keyword in ['float','int'] and node.condition:
                    env=self.env.copy()
                    env.autoref = node.name
                    with LogicalSolver(env) as s:
                        if not s.solve(node.condition):
                            raise Exception("Node does not fullfil a condition:",
                                            node.name, node.condition)
                # Check formats if set for strings
                if node.keyword=='str' and node.format:
                    m = re.match(node.format, node.value.value)
                    if not m:
                        raise Exception("Node value does not match the format:",
                                        node.value.value, node.format)
        return self.env
        
    def _determine_node(self, line):
        # Add replacement marks
        # TODO: we need to also properly treate arrays like this ["d#", "b"]
        encode = ["\\'", '\\"', "\n"]
        for i,symbol in enumerate(encode):
            line['code'] = line['code'].replace(symbol,f"$@{i:02d}")
            
        # Determine node type
        parser = Parser(
            code=line['code'],
            source=line['source']
        )
        steps = [
            EmptyNode.is_node,            # parse empty line node
            parser.part_indent,           
            ImportNode.is_node,           # parse root import directive
            UnitNode.is_node,             # parse unit directive
            SourceNode.is_node,           # parse source directive
            CaseNode.is_node,             # parse case directive
            OptionNode.is_node,           # parse option setting
            ConstantNode.is_node,         # parse constant setting
            FormatNode.is_node,           # parse format setting
            TagsNode.is_node,             # parse node tags
            ConditionNode.is_node,        # parse condition setting
            parser.part_name,             
            GroupNode.is_node,            # parse group node
            ImportNode.is_node,           # parse group import directive
            ModNode.is_node,              # parse modification
            parser.part_type,             
            BooleanNode.is_node,          # parse boolean node
            IntegerNode.is_node,          # parse integer node
            FloatNode.is_node,            # parse float node
            StringNode.is_node,           # parse string node
            TableNode.is_node,            # parse table node
        ]
        node = None
        for step in steps:
            if step.__name__=='is_node':
                node = step(parser)
                if node:
                    node = node
            else:
                step()
            if node and parser.is_empty():
                break
        else:
            raise Exception(f"Incorrect format: {code}")
        
        # Convert symbols to original letters
        def decode_symbols(value):
            replace = ["\'", '\"', "\n"]
            if isinstance(value, (list, np.ndarray)):
                value = [decode_symbols(v) for v in value]
            elif value is None:
                return value
            else:
                for i,symbol in enumerate(replace):
                    value = value.replace(f"$@{i:02d}", symbol)
            return value
        
        # Remove replacement marks
        if node.value_fn:
            node.value_fn = decode_symbols(node.value_fn)
        if node.value_expr:
            node.value_expr = decode_symbols(node.value_expr)
        node.value_raw = decode_symbols(node.value_raw)
        node.code = decode_symbols(node.code)

        # Return proper node type
        return node



from dataclasses import dataclass

@dataclass
class Source:
    filename: str          # file from which the source was called
    lineno: int            # line from which the source was called
    primary: bool = True   # is this a primary or imported source



from pygments.lexer import RegexLexer, include, bygroups
from pygments.token import Token
    
class SchemaLexer(RegexLexer):
    name = 'DIP'
    tokens = {
        'root': [
            (r'<indent>',     Token.Indent),
            (r'<name>',       Token.Name),
            (r'<type>',       Token.Type),
            (r'<unit>',       Token.Unit),
            (r'<value>',      Token.String),
            (r'<path>',       Token.String),
            (r'\.\.\.',       Token.Text),
            (r'(\()("""\n|"|\')(<expression>)(\n"""|"|\')(\))',
              bygroups(Token.Text, Token.Expression, Token.Expression,
                       Token.Expression, Token.Text)),
            (r'(\()(<function>)(\))',
              bygroups(Token.Text, Token.Expression,  Token.Text)),
            (r'<[^ ]',        Token.Name.Tag, 'tag'),
            (r'{',            Token.Reference, 'reference'),
            (r'!(options|format|condition|constant|tags)',  Token.Keyword),
            (r'\$(source|unit)',     Token.Keyword),
            (r'\@(case|else|end)',     Token.Keyword),
            (r'( |=|<|>|\.\*|\*)',    Token.Text),
            (r'#[^\n]*',  Token.Comment),
            (r'[ ]*\n',   Token.Text), 
        ],
        'tag': [
            (r'[a-zA-Z0-9_-]+', Token.Name.Tag),
            (r'>',              Token.Name.Tag, '#pop'),
        ],
        'reference': [
            (r'[^}]+',          Token.Reference),
            (r'{',              Token.Reference, '#push'),
            (r'}\[',            Token.Reference, 'reference_slice'),
            (r'}',              Token.Reference, '#pop'),
        ],
        'reference_slice': [
            (r'[^\]]+',         Token.Slice),
            (r',',              Token.Reference),
            (r'\]',             Token.Reference, '#pop:2'),
        ],

    }



from .node_base import BaseNode

class FormatNode(BaseNode):
    keyword: str = 'format'

    @staticmethod
    def is_node(parser):
        parser.kwd_format()
        if parser.is_parsed('kwd_format'):
            parser.part_value()
            parser.part_comment()
            return FormatNode(parser)
            
    def parse(self, env):
        if env.nodes[-1].keyword!='str':
            raise Exception("Format can be set only to string nodes", env.nodes[-1].code)
        env.nodes[-1].format = self.value_raw
        return None



from .node_base import BaseNode

class ConstantNode(BaseNode):
    keyword: str = 'constant'

    @staticmethod
    def is_node(parser):
        parser.kwd_constant()
        if parser.is_parsed('kwd_constant'):
            parser.part_comment()
            return ConstantNode(parser)
            
    def parse(self, env):
        env.nodes[-1].constant = True
        return None



import numpy as np
import itertools
import matplotlib.pyplot as plt

quant = ['Np','B','Bm/BmW/BW','BV/BuV','BuA','BOhm','BSPL','BSIL','BSWL','PR/AR','W','V','A','Ohm','Pa','W/m2']
nquant = len(quant)
idquant = list(range(nquant))

conv = [
    ['Np','B'],
    ['B','PR/AR'],
    ['Np','PR/AR'],
    ['Bm/BmW/BW','W'],
    ['BV/BuV','V'],
    ['BuA','A'],
    ['BOhm','Ohm'],
    ['BSPL','Pa'],
    ['BSIL','W/m2'],
    ['BSWL','W'],
]
for i in range(len(conv)):
    conv[i][0] = quant.index(conv[i][0])
    conv[i][1] = quant.index(conv[i][1])
conv = np.array(conv)

fig, axes = plt.subplots(1,1,figsize=(5,5),tight_layout=True)

ax = axes
ax.set_xticks(idquant)
ax.set_yticks(idquant)
ax.xaxis.set_major_locator(plt.MaxNLocator(nquant))
ax.yaxis.set_major_locator(plt.MaxNLocator(nquant))
#ax.set_xticklabels(quant, rotation='vertical', fontsize=8)
#ax.set_yticklabels(quant, rotation='horizontal', fontsize=8)
ax.set_axisbelow(True)
ax.grid()

ax.scatter(conv[:,0], conv[:,1])
#for q1, q2 in itertools.product(idquant,idquant):
#    where
#    print(q1, q2)

plt.savefig("logarithmic_conversions.png")


import sys, os
import pytest
import numpy as np
sys.path.insert(0, 'src')

from scinumtools.dip import DIP
from scinumtools.dip.solvers import LogicalSolver
from scinumtools.dip.datatypes import BooleanType

def test_or_and():
    with LogicalSolver() as p:
        assert p.solve('true || true || true')    == True
        assert p.solve('false || true || false')  == True
        assert p.solve('false || false || false') == False
        assert p.solve('true && true && true')    == True
        assert p.solve('true && false && true')   == False
        assert p.solve('false && false && false') == False
        assert p.solve('true && true && true || false || false') == True
        assert p.solve('false || true && false && true || true') == True
        assert p.solve('false || false || true && false && true') == False

def test_parenthesis():
    with LogicalSolver() as p:
        assert p.solve('(true || false) && true && true') == True
        assert p.solve('true && (true && false) && true') == False
        assert p.solve('false || false || (true || false)') == True
        assert p.solve('false || (true || false) && true') == True
        assert p.solve('false || true && (false || false)') == False
        assert p.solve('false || ((false||true) || false) && (true||false)') == True

def test_compare_nodes():
    with DIP() as dip:
        dip.from_string("""
        dogs int = 23
    cats int = 44
    birds int = 23
    fish int = 12
    animal bool = true
        """)
        env = dip.parse()
    with LogicalSolver(env) as p:
        # node pair comparison
        assert p.solve('{?dogs} == {?cats}') == False
        assert p.solve('{?dogs} == {?birds}') == True
        assert p.solve('{?dogs} != {?cats}') == True
        assert p.solve('{?dogs} != {?birds}') == False
        assert p.solve('{?dogs} <= {?cats}') == True
        assert p.solve('{?dogs} <= {?birds}') == True
        assert p.solve('{?dogs} <= {?fish}') == False
        assert p.solve('{?dogs} >= {?cats}') == False
        assert p.solve('{?dogs} >= {?birds}') == True
        assert p.solve('{?dogs} >= {?fish}') == True
        assert p.solve('{?dogs} <  {?cats}') == True
        assert p.solve('{?dogs} <  {?fish}') == False
        assert p.solve('{?dogs} >  {?fish}') == True
        assert p.solve('{?dogs} >  {?cats}') == False
        # single bool node
        assert p.solve('{?animal}') == True
        assert p.solve('~{?animal}') == False   # negated value
        # is node defined
        assert p.solve('!{?dogs}') == True
        assert p.solve('!{?elefant}') == False
        assert p.solve('!{?elefant} == false') == True
        assert p.solve('~!{?elefant}') == True  # negated value

def test_compare_values():
    with DIP() as dip:
        dip.from_string("""
    weight float = 57.3 kg
        """)
        env = dip.parse()
    with LogicalSolver(env) as p:
        # comparison with the same units
        assert p.solve('{?weight} == 57.30 kg') == True
        assert p.solve('{?weight} == 57.31 kg') == False
        assert p.solve('{?weight} != 57.30 kg') == False
        assert p.solve('{?weight} != 57.31 kg') == True
        assert p.solve('{?weight} <= 57.30 kg') == True
        assert p.solve('{?weight} <= 60 kg') == True
        assert p.solve('{?weight} <= 50 kg') == False
        # comparison with different units
        assert p.solve('{?weight} >= 57300 g') == True
        assert p.solve('{?weight} >= 50000 g') == True
        assert p.solve('{?weight} >= 60000 g') == False
        assert p.solve('{?weight} > 50000 g') == True
        assert p.solve('{?weight} > 60000 g') == False
        # comparison without specifying units
        assert p.solve('{?weight} < 50') == False
        assert p.solve('{?weight} < 60') == True

def test_combination():
    with DIP() as dip:
        dip.from_string("""
    size float = 34 cm
    geometry int = 2
      = 1 # line
      = 2 # square
      = 3 # cube
    filled bool = true
    dimension int = 2
        """)
        env = dip.parse()
    with LogicalSolver(env) as p:
        assert p.solve("""
        {?size} > 30 cm 
        || ({?size} < 0.4 m || {?size} >= 34) 
        && ({?geometry} == 1 && {?geometry}<={?dimension})
        && {?filled}
        || ~!{?color}
        """) == True
        
if __name__ == "__main__":
    # Specify wich test to run
    test = sys.argv[1] if len(sys.argv)>1 else True

    # Loop through all tests
    for fn in dir(sys.modules[__name__]):
        if fn[:5]=='test_' and (test is True or test==fn[5:]):
            print(f"\nTesting: {fn}\n")
            locals()[fn]()



from typing import List
import numpy as np

from ...units import Unit, UnitEnvironment
from .node_base import BaseNode
from .node_select import SelectNode
from ..datatypes import IntegerType,FloatType
from ..solvers import NumericalSolver, FunctionSolver

class IntegerNode(BaseNode, SelectNode):
    keyword: str = 'int'
    value: int = None
    options: List = None
    value_expr: str = False
    tags: list = None
    dtype = int
    bits = 32

    def __init__(self, *args, **kwargs):
        self.options = []
        super().__init__(*args, **kwargs)
        
    @staticmethod
    def is_node(parser):
        if parser.keyword=='int':
             parser.part_dimension()
             parser.part_equal()
             if parser.is_parsed('part_equal'): # definition
                 parser.part_value()  
             else:
                 parser.defined = True  # declaration
             parser.part_units()    
             parser.part_comment()
             return IntegerNode(parser)
         
    def set_value(self, value=None):
        """ Set value using value_raw or arbitrary value
        """
        if value is None and self.value_raw:
            self.value = IntegerType(self.cast_value(), self.units_raw)
        elif value:
            self.value = IntegerType(value, self.units_raw)
        else:
            self.value = None
            
    def parse(self, env):
        if self.value_fn: # Process function
            with FunctionSolver(env) as s:
                self.value_raw = s.solve(self.value_fn, self.units_raw)
        if self.value_expr: # Process expression
            with NumericalSolver(env) as s:
                self.value_raw = np.round(s.solve(self.value_expr, self.units_raw))
        # Testing validity of units
        if self.units_raw:
            with UnitEnvironment(env.units):
                Unit(self.units_raw)
        return None    



from .node_base import BaseNode

class EmptyNode(BaseNode):
    keyword: str = 'empty'

    @staticmethod
    def is_node(parser):
        if parser.is_empty():
            return EmptyNode(parser)
        parser.part_comment()
        if parser.is_parsed('part_comment'):
            return EmptyNode(parser)



import sys, os
import pytest
import numpy as np
sys.path.insert(0, 'src')

from scinumtools.dip import DIP
from scinumtools.dip.settings import Format
from scinumtools.dip.datatypes import FloatType, IntegerType, StringType, BooleanType

def parse(code):
    with DIP() as p:
        p.from_string(code)
        return p.parse().data(verbose=True,format=Format.TYPE)

def test_logical():
    data = parse('''
    a bool = true
    b float = 23.43 cm
    c bool = ("""
      false || {?b} == 23.43 cm && {?a}
    """)
    ''')
    np.testing.assert_equal(data,{
        'a': BooleanType(True),
        'b': FloatType(23.43, 'cm'),
        'c': BooleanType(True),
    })

def test_numerical():
    data = parse('''
    a float = 14.24 mm
    b int = 220 cm
    c float = ("{?a} + {?b} + 10 m") cm
    d int = ("{?b} + 1 cm + 10 m + 1 nm") cm
    ''')
    np.testing.assert_equal(data,{
        'a': FloatType(14.24, 'mm'),
        'b': IntegerType(220, 'cm'),
        'c': FloatType(1221.424, 'cm'),
        'd': IntegerType(1221, 'cm'),
    })
    with pytest.raises(Exception) as e_info:
        parse('''
        a float = ("10 dm + 1 m") J
        ''')
    """xxx
    assert e_info.value.args[0] == "Expression result dimensions do not match node dimensions:"
    """
    assert e_info.value.args[0] == "Unsupported conversion between units:"

def test_template():
    data = parse('''
    a float[2] = [14.24,15.23] mm
    b str = ("a = {{?a}[0]:.3e}")
    ''')
    np.testing.assert_equal(data,{
        'a': FloatType([14.24, 15.23], 'mm'),
        'b': StringType('a = 1.424e+01'),
    })
    
    
if __name__ == "__main__":
    # Specify wich test to run
    test = sys.argv[1] if len(sys.argv)>1 else True

    # Loop through all tests
    for fn in dir(sys.modules[__name__]):
        if fn[:5]=='test_' and (test is True or test==fn[5:]):
            print(f"\nTesting: {fn}\n")
            locals()[fn]()
    
